/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE TVPAINT ANIMATION LOADER/SAVER PLUGINS  *
 * SOURCE CODE.  													*
 *																	*
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A FREEBSD SOURCE LICENSE INCLUDED WITH THIS SOURCE   *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE TVPAINT ANIMATION LOADER/SAVER PLUGINS SOURCE CODE IS (C)    *
 * COPYRIGHT 2010-2015 BY THE TVPAINT DEVELOPPEMENT COMPANY         *
 * http://www.tvpaint.com/                                          *
 *                                                                  *
 ********************************************************************/

// ==================================================================
/** @file Format.SaverFFmpeg.cpp
      @brief File: FFmpeg container.
     @author Michael Schreiner & Eric SCHOLL & Xavier Richez
      $Date: $
        $Id: $
  @copyright (c) 2010-21 TVPaint Developpement. All Rights Reserved.
*/
// ==================================================================

// Mainly use this example
// http://ffmpeg.org/doxygen/trunk/doc_2examples_2muxing_8c-example.html

#include <string>
#ifndef __FORMAT_SAVERFFMPEG_HH
#include "Format.SaverFFmpeg.h"
#endif
#include "libavutil/error.h"

#if defined( WIN32 ) || defined( WIN64 )
	#define CAST_AVRATIONAL
#else
	#define CAST_AVRATIONAL		(AVRational)
#endif


enum
{
    kLogBufferChunkSize = 16384
};


cSaverFormatFFmpeg::~cSaverFormatFFmpeg()
{
}

cSaverFormatFFmpeg::cSaverFormatFFmpeg( eFormat iFormat ):
	mSrcAudioTrack( 0 ),
	mSrcVideoTrack( 0 ),
	mOption( 0 ),
	mFormat( iFormat ),
	//---
	mVideoStream( NULL ),
	mAudioStream( NULL ),
	mOutputFormat( NULL ),
	mFormatCtx( NULL ),
	//---
	mVideoFrame( NULL ),
	mVideoSrcPicture(),
	mVideoDstPicture(),
	mVideoSwsContext( 0 ),
	//---
	mAudioSrcSamplesData( 0 ),
	mAudioSrcSamplesLinesize( 0 ),
	mAudioSrcSamplesCount( 0 ),
	mAudioSwrContext( 0 ),
	mAudioFrame( 0 ),
	mAudioDstSamplesCountMax( 0 ),
	mAudioDstSamplesData( 0 ),
	mAudioDstSamplesLinesize( 0 ),
	mAudioDstSamplesSize( 0 ),
	mAudioDstSamplesCount( 0 ),
	mFirstSample( 0 ),
	mStartSample( 0 ),
	//---
	mNbPass( 1 ),
	mCurrentPass( 0 ),
	mCurrentFrame( 0 ),
	mLogBuffer( 0 ),
	mLogBufferSize( 0 )
{
    memset( mErrorMsg, 0, sizeof( mErrorMsg ) );
    av_register_all();

#ifdef DEBUG
	av_log_set_level( AV_LOG_TRACE );
#endif
}

//---

const char*
cSaverFormatFFmpeg::FormatName()
{
	switch( mFormat )
	{
	case kFormatMP4:
		return "MP4";
	case kFormatQT:
		return "MOV";

	case kFormatMKV:
		return "MKV";

#if 0 //TODO WebP
	case kFormatWebP:
		return "WebP";
#endif

	case kFormatWebM:
		return "WebM";
	default:
		return "ERROR-SAVER-NAME";
	}
}

const char*
cSaverFormatFFmpeg::Extension()
{
	switch( mFormat )
	{
	case kFormatMP4:
		return "mp4";
	case kFormatQT:
		return "mov";
	case kFormatMKV:
		return "mkv";

#if 0 // TODO WebP
	case kFormatWebP:
		return "webp";
#endif
	case kFormatWebM:
		return "webm";

	default:
		return "ERROR-SAVER-EXT";
	}
}

const char*
cSaverFormatFFmpeg::Copyright()
{
    return "Copyright";
}

const char*
cSaverFormatFFmpeg::Options( int iIdx )
{
	switch (mFormat)
	{
		case kFormatMP4:
		{
			switch (iIdx)
			{
			case 2:
				return "Lossless: PNG (RGB-RGBA)";
			case 1:
				return "Lossy: MPEG-4 Visual (RGB)";
			case 0:
				return "Lossy: H.264 (RGB-RGBA)";
			default:
				return "";
			}
		}
		case kFormatQT:
		{
			switch (iIdx)
			{
			case 2:
				return "Lossless: PNG (RGB-RGBA)";
			case 1:
				return "Lossy: MPEG-4 Visual (RGB)";
			case 0:
				return "Lossy: PRORES (RGB-RGBA)";
			default:
				return "";
			}
		}
		case kFormatMKV:
		{
			switch (iIdx)
			{
			case 2:
				return "Lossless: PNG (RGB-RGBA)";
			case 1:
				return "Lossy: MPEG-4 Visual (RGB)";
			case 0:
				return "Lossy: H.264 (RGB-RGBA)";
			default:
				return "";
			}
		}

#if 0 // TODO WebP
		case kFormatWebP:
		{
			switch (iIdx)
			{
			case 0:
				return "Lossy: VP9 (RGB-RGBA)";
			case 1:
				return "Lossy: VP8 (RGB-RGBA)";
			default:
				return "";
			}
		}
#endif
		case kFormatWebM:
		{
			switch (iIdx)
			{
			case 0:
#if !defined( WIN32 ) && !defined( WIN64 )
				return "Lossy: VP9 (RGB-RGBA)";
			case 1:
#endif
				return "Lossy: VP8 (RGB-RGBA)";
			default:
				return "";
			}
		}
		default:
			return "";
	}
}

void
cSaverFormatFFmpeg::Option( eSaverMode iIdx ) //Select option	//TODO: should be 'int' or an eOption enum (which will be output of the Options()) ?
{
	mOption = iIdx;
}

const char*
cSaverFormatFFmpeg::Version()
{
	return LOADERSAVER_VERSION_STR " ( " LIBAVFORMAT_IDENT " / " LIBAVCODEC_IDENT " )";
}

bool
cSaverFormatFFmpeg::UseBitrate()
{
	return true;
}

int
cSaverFormatFFmpeg::Bitrate( int iCodec, int iMaxBitRate )
{
	// slow but efficient
	if( iCodec == AV_CODEC_ID_VP9 )
		return iMaxBitRate/3;
	else if ( iCodec == AV_CODEC_ID_H264 )
		return iMaxBitRate/2;

	// good
	else if( iCodec == AV_CODEC_ID_VP8 || iCodec == AV_CODEC_ID_MPEG4 )
		return iMaxBitRate*2/3;

	// for old codecs
	else
		return iMaxBitRate;
}

//==================================================================================================================================================


//TODO: free memory

int
cSaverFormatFFmpeg::Save(const char* iPathFile, cSaverFrameSequence* iFrameSequence, cSaverAudioSequence* iAudioSequence, int iOptionIdx, int iBitRate)
{
    //Init
	mSrcVideoTrack = iFrameSequence;
	mSrcAudioTrack = iAudioSequence;
	mOption = iOptionIdx;

	cPixelBlockFF* block = 0;

    for( mCurrentPass = 1; mCurrentPass <= mNbPass; mCurrentPass++ )
    {
        //Init Format
		if( !InitFormat( iPathFile ) )
        {
            CloseFFmpeg();
            return -1;
        }

		int w = mSrcVideoTrack->Width();
		int h = mSrcVideoTrack->Height();

		if (mOutputFormat->video_codec == AV_CODEC_ID_H264 && w * h > (36864 /*MAX_MBS_PER_FRAME*/ << 8))
		{
			//limitation of openh264, see codec/encoder/core/src/encoder_ext.cpp
			SaveErrorString("Resolution too high for H264 codec");
			CloseFFmpeg();
			remove(iPathFile);
			return -1;
		}

        //Init Stream
		if( !InitVideoStream() )
		{
			CloseFFmpeg();
			return -1;
		}
        if( mNbPass == mCurrentPass )
        {
            if( !InitAudioStream() )
            {
                CloseFFmpeg();
                return -1;
            }
        }

        //Init Frame
		if( !InitVideoFrame() )
		{
			CloseFFmpeg();
			return -1;
		}
		if( mNbPass == mCurrentPass )
		{
			if( !InitAudioFrame() )
			{
				CloseFFmpeg();
				return -1;
			}
		}

        //Open File
        if( mNbPass == mCurrentPass )
		{
			if( !( mOutputFormat->flags & AVFMT_NOFILE ) )
			{
				int errCode = avio_open( &mFormatCtx->pb, iPathFile, AVIO_FLAG_WRITE );
				if( errCode < 0 )
				{
					SaveErrorString(errCode, "avio_open");
					CloseFFmpeg();
					return -1;
				}
			}

            //Write Header
            int errCode = avformat_write_header( mFormatCtx, NULL );
            if( errCode < 0 )
            {
                SaveErrorString(errCode, "avformat_write_header");
                CloseFFmpeg();
                remove( iPathFile );
                return -1;
            }
        }

        block = new cPixelBlockFF( w, h, new cPixel32[ w * h ] );
        if( !block->Pixels())
        {
            SaveErrorString("Cannot allocate pixels");
            CloseFFmpeg();
            remove( iPathFile );
            return -1;
        }

        int i;
        for( i = 0; i < mSrcVideoTrack->FrameCount(); i++ )
        {
            fflush(stdout);

            //Read Video
            if( !mSrcVideoTrack->ReadFrame( i, block ) )
            {
                SaveErrorString("Error reading video frame");
                CloseFFmpeg();
                remove( iPathFile );
                delete[] block->Pixels();
                delete block;
                return -1;
            }

            //Write Video
			if( !WriteVideoFrame( block, i ) )
            {
                CloseFFmpeg();
                remove( iPathFile );
                delete[] block->Pixels();
                delete block;
                return -1;
            }

            if( mNbPass == mCurrentPass )
            {
				if( !WriteAudioFrame( i ) )
                {
                    CloseFFmpeg();
                    remove( iPathFile );
                    delete[] block->Pixels();
                    delete block;
                    return -1;
                }
            }

            mSrcVideoTrack->Progress( ( mSrcVideoTrack->FrameCount() / mNbPass * ( mCurrentPass - 1 ) ) + i / mNbPass );

			if( mSrcVideoTrack->BreakSignal() )
			{
				SaveErrorString("Received abort signal");
				CloseFFmpeg();
				remove( iPathFile );
				delete[] block->Pixels();
				delete block;
				return -1;
			}
        }

		FlushFrames();
		if( mNbPass == mCurrentPass )
			FlushSamples();

        if( mNbPass != 1 && mCurrentPass == 1 )
        {
            CloseFFmpeg();
            delete[] block->Pixels();
            delete block;
        }
		else
		{
			delete[] mLogBuffer;
			mLogBuffer = 0;
			mLogBufferSize = 0;
		}
    }

	int errCode = av_write_trailer( mFormatCtx );
	//char  errchar[1001];
	//av_strerror( error, errchar, 1000 );
	if ( errCode < 0 )
	{
		SaveErrorString( errCode, "av_write_trailer" );
		CloseFFmpeg();
		remove( iPathFile );
		delete[] block->Pixels();
		delete block;
		return -1;
	}

    CloseFFmpeg();
    delete[] block->Pixels();
    delete block;

    return  0;
}

/*!    \brief Set the output format file, also avoid repeating theses lines code
 *    \param iFormat an extension file
 *    \param iPathFile path to the file that will be saved
 *    \param iShortName the name of the format that will be guessed by FFMPEG
 *    \return true if succeed, false otherwise
 */
bool cSaverFormatFFmpeg::SetOutputFormat( const char* iFormat, const char* iPathFile, const char* iShortName )
{
    int errCode = 0;
    if(iShortName!=0 )
        errCode = avformat_alloc_output_context2( &mFormatCtx, av_guess_format( iShortName, iPathFile, NULL ), iFormat, iPathFile );
    else
        errCode = avformat_alloc_output_context2( &mFormatCtx, 0, iFormat, iPathFile );
    if (errCode < 0)
    {
        SaveErrorString(errCode, "avformat_alloc_output_context2");
        return false;
    }
    else if (!mFormatCtx)
    {
        SaveErrorString("format allocation error");
        return false;
    }
    mOutputFormat = mFormatCtx->oformat;
    return true;
}

// To list all available codecs (with id/name/...)
// https://www.ffmpeg.org/doxygen/3.1/allcodecs_8c_source.html#l00061
// and/or search for ff_xxxxx_encoder/ff_xxxxxx_...

bool
cSaverFormatFFmpeg::InitFormat( const char* iPathFile )
{
	switch( mFormat )
	{
		case kFormatMP4:
		{

			if (!SetOutputFormat( "mp4", iPathFile, 0 ) )
				return false;
			switch( mOption )
			{
				case 2: //Lossless
					mOutputFormat->video_codec = AV_CODEC_ID_PNG;
					break;
				case 1:
					mOutputFormat->video_codec = AV_CODEC_ID_MPEG4;
					break;
				case 0:
					mOutputFormat->video_codec = AV_CODEC_ID_H264;
					break;
				default:
					SaveErrorString("Invalid MP4/codec combination !!\n");
					return false;
			}
			mOutputFormat->audio_codec = AV_CODEC_ID_AAC;	// ALAC not supported with mp4 container ... https://trac.ffmpeg.org/wiki/Encode/HighQualityAudio#Containerformats
		}
		break;

		case kFormatQT:
		{
			if (!SetOutputFormat( "mov", iPathFile, 0 ) )
				return false;

			switch( mOption )
			{
				case 2: //Lossless
					mOutputFormat->video_codec = AV_CODEC_ID_PNG;
					break;
				case 1:
					mOutputFormat->video_codec = AV_CODEC_ID_MPEG4;
					break;
				case 0:
					mOutputFormat->video_codec = AV_CODEC_ID_PRORES;
					break;
				default:
					SaveErrorString( "Invalid QT/codec combination !!\n" );
					return false;
			}
		mOutputFormat->audio_codec = AV_CODEC_ID_ALAC; //Apple Lossless Audio Codec
		}
		break;

		case kFormatMKV:
		{
			if (!SetOutputFormat( "mkv", iPathFile, "matroska" ) )
				return false;

			switch (mOption)
			{
				case 2: //Lossless
					mOutputFormat->video_codec = AV_CODEC_ID_PNG;
					break;
				case 1:
					mOutputFormat->video_codec = AV_CODEC_ID_MPEG4;
					break;
				case 0:
					mOutputFormat->video_codec = AV_CODEC_ID_H264;
					break;
				default:
					SaveErrorString( "Invalid MKV/codec combination !!\n" );
					return false;
			}
			mOutputFormat->audio_codec = AV_CODEC_ID_AAC;

			// Also known as VOBSUB, basic subtitles format, easy to edit and small size, for better quality see AV_CODEC_ID_HDMV_PGS_SUBTITLE
			mOutputFormat->subtitle_codec = AV_CODEC_ID_DVD_SUBTITLE;
		}
		break;
#if 0 // TODO WebP
		case kFormatWebP:
		{
			if (!SetOutputFormat( "webp", iPathFile, "webp" ))
				return false;

			switch (mOption)
			{
				case 0:
					mOutputFormat->video_codec = AV_CODEC_ID_VP9;
					break;
				case 1:
					mOutputFormat->video_codec = AV_CODEC_ID_VP8;
					break;
				default:
					SaveErrorString( "Invalid WebP/codec combination !!\n" );
					return false;
			}
		}
		break;
#endif
		case kFormatWebM:
		{
			if (!SetOutputFormat( "webm", iPathFile, "webm" ))
				return false;
#if defined( WIN32 ) || defined( WIN64 )
			mOutputFormat->video_codec = AV_CODEC_ID_VP8;
#else
			switch (mOption)
			{
				case 0:
					mOutputFormat->video_codec = AV_CODEC_ID_VP9;
					break;
				case 1:
					mOutputFormat->video_codec = AV_CODEC_ID_VP8;
					break;
				default:
					SaveErrorString( "Invalid WebM/codec combination !!\n" );
					return false;
			}
#endif
			mOutputFormat->audio_codec = AV_CODEC_ID_OPUS;
		}
		break;
		default:
			return false;
	}

	return true;
}

bool
cSaverFormatFFmpeg::InitVideoStream()
{
	AVCodec* codec = 0;		// don't destroy, owned by FFMPEG
	if( mOutputFormat->video_codec == AV_CODEC_ID_PRORES )
		codec = avcodec_find_encoder_by_name( "prores_ks" );
	else
		codec = avcodec_find_encoder( mOutputFormat->video_codec );

	if( !codec )
	{
		std::string error = "Video codec not found";
		switch( mOutputFormat->video_codec )
		{
			case AV_CODEC_ID_MPEG4:
				error += ": MPEG.4";
				break;
			case AV_CODEC_ID_H264:
				error += ": H.264";
				break;
			case AV_CODEC_ID_QTRLE:
				error += ": QT RLE";
				break;
			case AV_CODEC_ID_PNG:
				error += ": PNG";
				break;
			case AV_CODEC_ID_PRORES:
				error += ": ProRes";
				break;
			case AV_CODEC_ID_VP9:
				error += ": VP9";
				break;
			case AV_CODEC_ID_VP8:
				error += ": VP8";
				break;
			default:
				break;
		}
		SaveErrorString( error.c_str() );
		return false;
	}

	mVideoStream = avformat_new_stream( mFormatCtx, codec );
	if( !mVideoStream )
	{
		SaveErrorString("Couldn't create video stream");
		return false;
	}

	mVideoStream->id = mFormatCtx->nb_streams - 1;
	mVideoStream->sample_aspect_ratio = av_d2q( mSrcVideoTrack->mPixelAspectRatio, 255 );

	bool ret = InitVideoCodecContext( mVideoStream->codec );

	mVideoStream->time_base = mVideoStream->codec->time_base;

	return ret;
}

bool
cSaverFormatFFmpeg::InitAudioStream()
{
    if( !mSrcAudioTrack )
		return true;

	AVCodec* codec = avcodec_find_encoder( mOutputFormat->audio_codec ); // don't destroy, owned by FFMPEG
	if( !codec )
	{
		SaveErrorString("Audio codec not found");
		return false;
	}

	mAudioStream = avformat_new_stream( mFormatCtx, codec );
	if( !mAudioStream )
	{
		SaveErrorString("Couldn't create audio stream");
		return false;
	}

	mAudioStream->id = mFormatCtx->nb_streams - 1;

	bool ret = InitAudioCodecContext( mAudioStream->codec );

	mAudioStream->time_base = mAudioStream->codec->time_base;

	return ret;
}

bool
cSaverFormatFFmpeg::InitVideoCodecContext( AVCodecContext* ioCodecContext )
{
	const AVCodec* codec = ioCodecContext->codec;

	uint64_t maxBitRate = uint64_t( mSrcVideoTrack->Width() * mSrcVideoTrack->Height() * mSrcVideoTrack->FrameRate() * 3 /*rgb*/ * 8 /*bits*/ ); //arbitrary
	if( maxBitRate > 240000000 /* ~= 30Mo/s */ )
		maxBitRate = 240000000;
	ioCodecContext->bit_rate = Bitrate( codec->id, maxBitRate );

	// To force the bitrate of the the output file, but sometimes it doesn't work with too low resolution or ... missing other parameters or ... I don't know -_-
	//ioCodecContext->rc_max_rate = ioCodecContext->bit_rate;
	//ioCodecContext->rc_min_rate = ioCodecContext->bit_rate;
//TODO:2.8?    //ioCodecContext->bit_rate_tolerance = ioCodecContext->bit_rate / 4;
//TODO:2.8?    //ioCodecContext->qmin = 1;
//TODO:2.8?    //ioCodecContext->qmax = 51;

    //define codec type
	ioCodecContext->codec_id = codec->id;
	ioCodecContext->codec_type = AVMEDIA_TYPE_VIDEO;

    //define sizes
	ioCodecContext->width = mSrcVideoTrack->Width();
	ioCodecContext->height = mSrcVideoTrack->Height();

	ioCodecContext->pix_fmt = GetPixelFormat( codec );
	ioCodecContext->sample_aspect_ratio = av_d2q( mSrcVideoTrack->mPixelAspectRatio, 255 );
////ZAP:	iCodecContext->me_method = ME_EPZS;
//TODO:2.8?       //ioCodecContext->rc_buffer_size = ioCodecContext->width * ioCodecContext->height * 4;
//TODO:2.8?       //ioCodecContext->rc_initial_buffer_occupancy = ioCodecContext->rc_buffer_size * 3 / 4;
//TODO:2.8?    ioCodecContext->rc_buffer_aggressivity = 1.0;
	ioCodecContext->time_base = CAST_AVRATIONAL{ 1000, int( mSrcVideoTrack->FrameRate() * 1000 ) };
	av_reduce( &ioCodecContext->time_base.num, &ioCodecContext->time_base.den, ioCodecContext->time_base.num, ioCodecContext->time_base.den, 90000 );

	if( mFormatCtx->oformat->flags & AVFMT_GLOBALHEADER )
		ioCodecContext->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;

	switch( codec->id )
    {
		case AV_CODEC_ID_MPEG4:
		case AV_CODEC_ID_H264:
		    ioCodecContext->gop_size = 10;
		    mNbPass = 2;
		    break;
		case AV_CODEC_ID_QTRLE:
		case AV_CODEC_ID_PNG:
		case AV_CODEC_ID_PRORES:
		case AV_CODEC_ID_VP9:
		case AV_CODEC_ID_VP8:
		    ioCodecContext->gop_size = 0;
		    break;
		default:
		    return false;
    }

	mVideoStream->codec->stats_in = 0;
    if( mNbPass != 1 )
    {
        if( mCurrentPass == 1 )
        {
            mLogBuffer = (char*)malloc( kLogBufferChunkSize );
            mLogBuffer[0] = 0;
            mLogBufferSize = kLogBufferChunkSize;
			ioCodecContext->flags |= AV_CODEC_FLAG_PASS1;
        }
        else
        {
			ioCodecContext->flags |= AV_CODEC_FLAG_PASS2;
            mVideoStream->codec->stats_in = mLogBuffer;
        }
    }

    int errCode = avcodec_open2( ioCodecContext, codec, NULL );
	if( errCode < 0 )
	{
		SaveErrorString(errCode, "avcodec_open2");
        return false;
	}

	return true;
}

double
cSaverFormatFFmpeg::GetSamplesCountPerImage() const
{
	return ( 1.0 / mSrcVideoTrack->FrameRate() ) * mSrcAudioTrack->SamplingRate();
}

bool
cSaverFormatFFmpeg::InitAudioCodecContext( AVCodecContext* ioCodecContext )
{
	const AVCodec* codec = ioCodecContext->codec;

	ioCodecContext->codec_id = codec->id;
	ioCodecContext->codec_type = AVMEDIA_TYPE_AUDIO;

	if( mFormatCtx->oformat->flags & AVFMT_GLOBALHEADER )
		ioCodecContext->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;

	ioCodecContext->sample_fmt = GetSampleFormat( codec );
	ioCodecContext->sample_rate = mSrcAudioTrack->SamplingRate();
	if( codec->id == AV_CODEC_ID_OPUS )
	{
		if( ( ioCodecContext->sample_rate == 44100 ) ||
			( ioCodecContext->sample_rate == 22050 ) ||
			( ioCodecContext->sample_rate == 11025 ) )
		{
			SaveErrorString("Opus -the audio codec for WebM- accepts only these\nsound sample rates: 12000, 16000, 24000 or 48000 (48000 advised)");
			return false;
		}
	}
	ioCodecContext->channels = mSrcAudioTrack->ChannelCount();
	ioCodecContext->channel_layout = av_get_default_channel_layout( ioCodecContext->channels );
	ioCodecContext->bit_rate = mSrcAudioTrack->BytesPerSample() * mSrcAudioTrack->SamplingRate() * 8;
	if( (codec->id == AV_CODEC_ID_OPUS) && (ioCodecContext->bit_rate > 512000) )
		ioCodecContext->bit_rate = 512000;
	//ioCodecContext->bit_rate = 160000;									// MP2 value: https://www.ffmpeg.org/doxygen/2.1/mpegaudiodata_8c.html#ac7a51a6f7c5530cfbdd54dfa5cf0752b
	//ioCodecContext->strict_std_compliance = FF_COMPLIANCE_EXPERIMENTAL;	//needed to use some codecs
	ioCodecContext->time_base = CAST_AVRATIONAL{ 1, ioCodecContext->sample_rate };

	int errCode = avcodec_open2( ioCodecContext, codec, NULL );
	if( errCode < 0 )
	{
		SaveErrorString(errCode, "avcodec_open2");
        return false;
	}

	return true;
}

AVPixelFormat
cSaverFormatFFmpeg::GetPixelFormat( const AVCodec* iCodec )
{
	AVPixelFormat pformat = AV_PIX_FMT_NONE;

	for( int i = 0; ( iCodec->pix_fmts != NULL && iCodec->pix_fmts[i] != -1 ); i++ )
	{
		if( iCodec->pix_fmts[i] == AV_PIX_FMT_RGB32 || iCodec->pix_fmts[i] == AV_PIX_FMT_BGR32 || iCodec->pix_fmts[i] == AV_PIX_FMT_ABGR || iCodec->pix_fmts[i] == AV_PIX_FMT_ARGB || iCodec->pix_fmts[i] == AV_PIX_FMT_BGRA || iCodec->pix_fmts[i] == AV_PIX_FMT_RGBA )
		{
			pformat = iCodec->pix_fmts[i];
			break;
		}
		else if( iCodec->pix_fmts[i] == AV_PIX_FMT_RGB24 || iCodec->pix_fmts[i] == AV_PIX_FMT_BGR24 || iCodec->pix_fmts[i] == AV_PIX_FMT_PAL8 )
		{
			pformat = iCodec->pix_fmts[i];
		}
		else if( iCodec->pix_fmts[i] == AV_PIX_FMT_YUV420P || iCodec->pix_fmts[i] == AV_PIX_FMT_YUVJ420P )
		{
			pformat = iCodec->pix_fmts[i];
		}
		//else if( iCodec->pix_fmts[i] == AV_PIX_FMT_YUV422P10 )
		//else if( iCodec->pix_fmts[i] == AV_PIX_FMT_YUV444P10 )
		else if( iCodec->pix_fmts[i] == AV_PIX_FMT_YUVA444P10 )
		{
			pformat = iCodec->pix_fmts[i];
			break;
		}
	}

	if( pformat == AV_PIX_FMT_NONE && iCodec->pix_fmts != NULL )
	{
		pformat = iCodec->pix_fmts[0];
	}

	return pformat;
}

AVSampleFormat
cSaverFormatFFmpeg::GetSampleFormat( const AVCodec* iCodec )
{
	AVSampleFormat pformat = AV_SAMPLE_FMT_NONE;

	for( int i = 0; ( iCodec->sample_fmts != NULL && iCodec->sample_fmts[i] != -1 ); i++ )
	{
		if( iCodec->sample_fmts[i] == AV_SAMPLE_FMT_S16 )
		{
			pformat = iCodec->sample_fmts[i];
			break;
		}
	}

	if( pformat == AV_SAMPLE_FMT_NONE && iCodec->sample_fmts != NULL )
	{
		pformat = iCodec->sample_fmts[0];
	}

	return pformat;
}

static
int
WritePacket( AVFormatContext* iFormatCtx, const AVRational* iTimeBase, const AVStream* iStream, AVPacket* ioPacket )
{
	ioPacket->pts = av_rescale_q_rnd( ioPacket->pts, *iTimeBase, iStream->time_base, AVRounding( AV_ROUND_NEAR_INF | AV_ROUND_PASS_MINMAX ) );
	ioPacket->dts = av_rescale_q_rnd( ioPacket->dts, *iTimeBase, iStream->time_base, AVRounding( AV_ROUND_NEAR_INF | AV_ROUND_PASS_MINMAX ) );
	ioPacket->duration = av_rescale_q( ioPacket->duration, *iTimeBase, iStream->time_base );
	ioPacket->stream_index = iStream->index;

	return av_interleaved_write_frame( iFormatCtx, ioPacket );
}

bool
cSaverFormatFFmpeg::InitVideoFrame()
{
	mVideoFrame = av_frame_alloc();
	if( !mVideoFrame )
	{
		SaveErrorString("Couldn't allocate video frame");
		return false;
	}

	mVideoFrame->format = mVideoStream->codec->pix_fmt;
	mVideoFrame->width = mVideoStream->codec->width;
	mVideoFrame->height = mVideoStream->codec->height;

	int errCode = avpicture_alloc( &mVideoDstPicture, mVideoStream->codec->pix_fmt, mVideoStream->codec->width, mVideoStream->codec->height );
	if ( errCode < 0 )
	{
		SaveErrorString(errCode, "avpicture_alloc");
		return false;
	}

	// mVideoSrcPicture is created/initialized with avpicture_fill() using the iBlock

	*( (AVPicture *)mVideoFrame ) = mVideoDstPicture;

	//---

	mVideoSwsContext = sws_getContext( mVideoStream->codec->width, mVideoStream->codec->height, AV_PIX_FMT_BGRA,
										mVideoStream->codec->width, mVideoStream->codec->height, mVideoStream->codec->pix_fmt,
										SWS_BICUBIC, NULL, NULL, NULL );
	if( !mVideoSwsContext )
	{
		SaveErrorString("Couldn't get context");
		return false;
	}

	return true;
}

bool
cSaverFormatFFmpeg::InitAudioFrame()
{
	if( !mSrcAudioTrack )
		return true;

	mFirstSample = int( mSrcAudioTrack->Offset() * mSrcVideoTrack->FrameRate() * GetSamplesCountPerImage() ) + 1;
	mStartSample = mFirstSample;

	mAudioSrcSamplesCount = mAudioStream->codec->frame_size;
	int errCode = av_samples_alloc_array_and_samples( &mAudioSrcSamplesData, &mAudioSrcSamplesLinesize, mAudioStream->codec->channels, mAudioSrcSamplesCount, AV_SAMPLE_FMT_S16, 0 );
	if ( errCode < 0 )
	{
		SaveErrorString(errCode, "av_samples_alloc_array_and_samples");
		return false;
	}

	//---

	mAudioDstSamplesCountMax = mAudioSrcSamplesCount;
	mAudioDstSamplesCount = 0;

	mAudioSwrContext = swr_alloc();
	if( !mAudioSwrContext )
	{
		SaveErrorString("SWR alloc failed");
		return false;
	}

	av_opt_set_int( mAudioSwrContext, "in_channel_count", mAudioStream->codec->channels, 0 );	// Set from TVPaint
	av_opt_set_int( mAudioSwrContext, "in_sample_rate", mAudioStream->codec->sample_rate, 0 );	// Set from TVPaint
	av_opt_set_sample_fmt( mAudioSwrContext, "in_sample_fmt", AV_SAMPLE_FMT_S16, 0 );
	av_opt_set_int( mAudioSwrContext, "out_channel_count", mAudioStream->codec->channels, 0 );	// Set from TVPaint
	av_opt_set_int( mAudioSwrContext, "out_sample_rate", mAudioStream->codec->sample_rate, 0 );	// Set from TVPaint
	av_opt_set_sample_fmt( mAudioSwrContext, "out_sample_fmt", mAudioStream->codec->sample_fmt, 0 );

	errCode =  swr_init( mAudioSwrContext );
	if ( errCode < 0 )
	{
		SaveErrorString(errCode, "swr_init");
		return false;
	}

	errCode =  av_samples_alloc_array_and_samples( &mAudioDstSamplesData, &mAudioDstSamplesLinesize, mAudioStream->codec->channels, mAudioDstSamplesCountMax, mAudioStream->codec->sample_fmt, 0 );
	if ( errCode < 0 )
	{
		SaveErrorString(errCode, "av_samples_alloc_array_and_samples");
		return false;
	}

	mAudioDstSamplesSize = av_samples_get_buffer_size( 0, mAudioStream->codec->channels, mAudioDstSamplesCountMax, mAudioStream->codec->sample_fmt, 0 );
	if( mAudioDstSamplesSize < 0 )
	{
		SaveErrorString( mAudioDstSamplesSize, "av_samples_get_buffer_size" );
		return false;
	}

	mAudioFrame = av_frame_alloc();
	if( !mAudioFrame )
	{
		SaveErrorString( "Could not allocate" );
		return false;
	}

	return true;
}

bool
cSaverFormatFFmpeg::WriteVideoFrame( cPixelBlockFF* iBlock, int iFrameNumber )
{
	int errCode =  avpicture_fill( &mVideoSrcPicture, (const uint8_t*)iBlock->Pixels(), AV_PIX_FMT_BGRA, iBlock->Width(), iBlock->Height() );
	if ( errCode < 0 )
	{
		SaveErrorString(errCode, "avpicture_fill");
		return false;
	}

	sws_scale( mVideoSwsContext,
				mVideoSrcPicture.data, mVideoSrcPicture.linesize,
				0, mVideoStream->codec->height,
				mVideoDstPicture.data, mVideoDstPicture.linesize );

	//---

	if( mFormatCtx->oformat->flags & FF_API_AVPICTURE )
	{
		SaveErrorString("Incompatible video flags");
		return false;
	}

	//---

    mVideoFrame->pts = iFrameNumber;

	AVPacket packet = { 0 };
	av_init_packet( &packet );

	int got_packet = 0;
	errCode =  avcodec_encode_video2( mVideoStream->codec, &packet, mVideoFrame, &got_packet );
	if ( errCode < 0 )
	{
		SaveErrorString( errCode, "avcodec_encode_video2");
		return false;
	}

    if( mNbPass == mCurrentPass )
    {
		if( got_packet )
		{
			int errCode =  WritePacket( mFormatCtx, &mVideoStream->codec->time_base, mVideoStream, &packet );
			if ( errCode < 0 )
			{
				SaveErrorString(errCode, "av_interleaved_write_frame");
				return false;
			}
		}

        avcodec_flush_buffers(mVideoStream->codec);
    }
    else
	{
		if( got_packet )
			av_free_packet( &packet );

		if( mVideoStream->codec->stats_out )
		{
			size_t s = strlen( mVideoStream->codec->stats_out );
			if( strlen( mLogBuffer ) + s >= mLogBufferSize )
			{
				mLogBufferSize += kLogBufferChunkSize;
				mLogBuffer = (char*)realloc( mLogBuffer, mLogBufferSize );
			}
#if defined( WIN32 ) || defined( WIN64 )
			strcat_s( mLogBuffer, mLogBufferSize, mVideoStream->codec->stats_out );
#else
			strcat( mLogBuffer, mVideoStream->codec->stats_out );
#endif
		}
    }

	return true;
}


// Must be done like this (for each frame), because FFmpeg need to write image/sound/image/sound/... to have an easily readable file
bool
cSaverFormatFFmpeg::WriteAudioFrame( int iFrameNumber )
{
	if( !mSrcAudioTrack )
		return true;

	int current_sample = mStartSample;
	int end_sample_in_next_image = mFirstSample + int( ( iFrameNumber + 1 ) * GetSamplesCountPerImage() ) + 1;
	while( current_sample < end_sample_in_next_image )
	{
		int src_sample_read = 0;
		if( mSrcAudioTrack->ReadData( (int16_t*)mAudioSrcSamplesData[0], mAudioSrcSamplesCount, &src_sample_read, current_sample ) )
		{
			SaveErrorString("Error reading audio data");
			return false;
		}

		if( src_sample_read <= 0 )	// silence gap
		{
			//break; //No more to read
			src_sample_read = mAudioSrcSamplesCount;
			av_samples_set_silence( mAudioSrcSamplesData, 0, mAudioSrcSamplesCount, mAudioStream->codec->channels, AV_SAMPLE_FMT_S16 );
		}

		//---

		int dst_nb_samples = av_rescale_rnd( swr_get_delay( mAudioSwrContext, mAudioStream->codec->sample_rate ) + mAudioSrcSamplesCount, mAudioStream->codec->sample_rate, mAudioStream->codec->sample_rate, AV_ROUND_UP );
		if( dst_nb_samples > mAudioDstSamplesCountMax )
		{
			av_free( mAudioDstSamplesData[0] );
			int errCode =  av_samples_alloc( mAudioDstSamplesData, &mAudioDstSamplesLinesize, mAudioStream->codec->channels, dst_nb_samples, mAudioStream->codec->sample_fmt, 0 );
			if ( errCode < 0 )
			{
				SaveErrorString(errCode, "av_samples_alloc");
				return false;
			}

			mAudioDstSamplesCountMax = dst_nb_samples;
			mAudioDstSamplesSize = av_samples_get_buffer_size( NULL, mAudioStream->codec->channels, dst_nb_samples, mAudioStream->codec->sample_fmt, 0 );
		}
		int errCode =  swr_convert( mAudioSwrContext, mAudioDstSamplesData, dst_nb_samples, (const uint8_t **)mAudioSrcSamplesData, src_sample_read );
		if ( errCode < 0 )
		{
			SaveErrorString(errCode, "swr_convert");
			return false;
		}

		//TODO: also flush swr_convert( ..., 0, 0 ) like for Loader ????????????????????????????

		//---

		mAudioFrame->nb_samples = dst_nb_samples;
		mAudioFrame->pts = av_rescale_q( mAudioDstSamplesCount, CAST_AVRATIONAL{ 1, mAudioStream->codec->sample_rate }, mAudioStream->codec->time_base );
		avcodec_fill_audio_frame( mAudioFrame, mAudioStream->codec->channels, mAudioStream->codec->sample_fmt, mAudioDstSamplesData[0], mAudioDstSamplesSize, 0 );
		mAudioDstSamplesCount += dst_nb_samples;

		//---

		AVPacket pkt = { 0 };
		av_init_packet( &pkt );

		int got_packet = 0;
		errCode =  avcodec_encode_audio2( mAudioStream->codec, &pkt, mAudioFrame, &got_packet );
		if ( errCode < 0 )
		{
			SaveErrorString(errCode, "avcodec_encode_audio2");
			return false;
		}

		if( !got_packet )
			break;

		errCode =  WritePacket( mFormatCtx, &mAudioStream->codec->time_base, mAudioStream, &pkt );
		if ( errCode < 0 )
		{
			SaveErrorString(errCode, "av_interleaved_write_frame");
			return false;
		}

		//---

		current_sample += src_sample_read;
	}

	mStartSample = current_sample;

	avcodec_flush_buffers( mAudioStream->codec );

    return true;
}

bool
cSaverFormatFFmpeg::FlushFrames()
{
	while( 1 )
	{
		AVPacket packet = { 0 };
		av_init_packet( &packet );

		int got_packet = 0;
		int errCode =  avcodec_encode_video2( mVideoStream->codec, &packet, 0, &got_packet );
		if ( errCode < 0 )
		{
			SaveErrorString( errCode, "avcodec_encode_video2");
			return false;
		}

		if( !got_packet )
			break;

		errCode =  WritePacket( mFormatCtx, &mVideoStream->codec->time_base, mVideoStream, &packet );
		if ( errCode < 0 )
		{
			SaveErrorString(errCode, "av_interleaved_write_frame");
			return false;
		}
	}

	return true;
}

bool
cSaverFormatFFmpeg::FlushSamples()
{
	if( !mSrcAudioTrack )
		return true;

	while( 1 )
	{
		AVPacket packet = { 0 };
		av_init_packet( &packet );

		int got_packet = 0;
		int errCode =  avcodec_encode_audio2( mAudioStream->codec, &packet, 0, &got_packet );
		if ( errCode < 0 )
		{
			SaveErrorString(errCode, "avcodec_encode_audio2");
			return false;
		}

		if( !got_packet )
			break;

		errCode =  WritePacket( mFormatCtx, &mAudioStream->codec->time_base, mAudioStream, &packet );
		if ( errCode < 0 )
		{
			SaveErrorString(errCode, "av_interleaved_write_frame");
			return false;
		}
	}

	return true;
}

void
cSaverFormatFFmpeg::CloseFFmpeg()
{
	// Close video codec/context
	if( mVideoStream )
    {
		avcodec_close( mVideoStream->codec );
		//av_free( mVideoSrcPicture.data[0] );	// not needed because it comes from the TVPaint block
		memset( &mVideoSrcPicture, 0, sizeof( mVideoSrcPicture ) );
		if( mVideoDstPicture.data && mVideoDstPicture.data[0] )
			av_free( mVideoDstPicture.data[0] );
		memset( &mVideoDstPicture, 0, sizeof( mVideoDstPicture ) );
		if( mVideoFrame )
		{
			av_frame_free( &mVideoFrame );
			mVideoFrame = 0;
		}
		if( mVideoSwsContext )
		{
			sws_freeContext( mVideoSwsContext );
			mVideoSwsContext = 0;
		}
	}

    // Close audio codec/buffer/context
	if( mAudioStream )
    {
		avcodec_close( mAudioStream->codec );
		if( mAudioDstSamplesData )
		{
			if( mAudioDstSamplesData[0] )
				av_free( mAudioDstSamplesData[0] );
			av_free( mAudioDstSamplesData );
			mAudioDstSamplesData = 0;
		}
		if( mAudioSrcSamplesData )
		{
			if( mAudioSrcSamplesData[0] )
				av_free( mAudioSrcSamplesData[0] );
			av_free( mAudioSrcSamplesData );
			mAudioSrcSamplesData = 0;
		}
		if( mAudioFrame )
		{
			av_frame_free( &mAudioFrame );
			mAudioFrame = 0;
		}
		if( mAudioSwrContext )
		{
			swr_free( &mAudioSwrContext );
			mAudioSwrContext = 0;
		}
	}

	if( mOutputFormat && !( mOutputFormat->flags & AVFMT_NOFILE ) )
		avio_close( mFormatCtx->pb );

	// Close format/stream
	avformat_free_context( mFormatCtx );
	mFormatCtx = 0;
	mOutputFormat = 0;
	mVideoStream = 0;	// deleted above
	mAudioStream = 0;	// deleted above
}

bool
cSaverFormatFFmpeg::Audio()
{
	return true;
}

uint8_t cSaverFormatFFmpeg::SaverMode()
{
	switch( mOption )
	{
	case 0:
	case 2:
		return ( bSaveRGB | bSaveRGBA );
	case 1:
	default:
		return bSaveRGB;
	}
}


void cSaverFormatFFmpeg::SaveErrorString(const char* msg)
{
	memset( mErrorMsg, 0, 256 );
	strncpy( mErrorMsg, msg, 256 );
}

void cSaverFormatFFmpeg::SaveErrorString(int code, const char* defaultMsg)
{
	memset( mErrorMsg, 0, 256 );
	if ( !av_strerror(code, mErrorMsg, 256) )
	{
		if( defaultMsg )
		{
			char av_error[256] = { 0 };
			sprintf( mErrorMsg, "Error %s: %d", defaultMsg, av_make_error_string(av_error, 256, code));
		}
		else
			strcpy( mErrorMsg, "Unknown FFmpeg error code" );
	}
}

const char* cSaverFormatFFmpeg::Error()
{
    if (mErrorMsg[0] != 0)
        return mErrorMsg;
    else
        return NULL;
}


